#include <memory>
#include <vector>
#include <string>
#include <utility>
#include <fstream>
#include <sstream>
#include <iostream>
#include <iterator>
#include <stdexcept>
#include <cstring>
#include <cerrno>
#include <cstdlib>
#include <sys/types.h>
#include <unistd.h>

using namespace std;


void waitKey(void)
{
  char c;
  cin.read(&c, 1);
}

size_t procMem(void)
{
  // file name
  stringstream path;
  path << "/proc/" << getpid() << "/statm";
  // read file to the memory
  ifstream pm( path.str() );
  if( !pm.is_open() )
  {
    stringstream err;
    err << "cannot open file '" << path.str() << "': " << strerror(errno);
    throw std::runtime_error( err.str() );
  }
  // parse
  vector<size_t> content{ istream_iterator<size_t>{pm}, istream_iterator<size_t>{} };
  const size_t pages    = content.at(5);    // i.e. data+stack
  const size_t pageSize = getpagesize();    // page is is required as well
  // return result
  return pages * pageSize;
}

int main(int argc, char **argv)
{
  if(argc!=1+2)
  {
    cerr << argv[0] << " <big_allocs> <keep_last>" << endl;
    cerr << "\tparameters are '1'/'0' (i.e. true/false)" << endl;
    return 2;
  }

  constexpr size_t mb = 1024*1024;

  const bool bigAllocs = atoi(argv[1]);
  const bool keepLast  = atoi(argv[2]);

  typedef unique_ptr<char[]> MemBlock;
  typedef vector<MemBlock>   MemPool;

  size_t blSize;
  size_t mpSize;
  if(bigAllocs)
  {
    blSize = 1024;
    mpSize = 1*1024*1024;
  }
  else
  {
    blSize = 8;
    mpSize = 128*1024*1024;
  }
  cout << "hello - using " << procMem()/mb << "MB of memory" << endl;

  MemPool mp;
  mp.reserve(mpSize);
  cout << "allocating " << mpSize << " blocks of " << blSize << " bytes (" << (mpSize*blSize)/mb << "MB total)" << endl;
  for(size_t i=0; i<mpSize; ++i)
  {
    MemBlock tmp(new char[blSize]);
    bzero(tmp.get(), blSize);
    mp.push_back( std::move(tmp) );
  }

  cout << "done - using " << procMem()/mb << "MB of memory; press enter to continue" << endl;
  waitKey();

  cout << "deallocationg all" << endl;
  MemBlock last;
  {
    if(keepLast)
      last = std::move( *mp.rbegin() );
    MemPool tmp;
    tmp.swap(mp);
  }
  cout << "done - using " << procMem()/mb << "MB of memory; press enter to exit" << endl;
  waitKey();

  return 0;
}
